use patina::platform;

mod convert;
mod render;

mod prelude {
    pub use crate::convert::{FromSDL,ToSDL};
}

pub struct SDLInitFlags {
    pub window_width: usize,
    pub window_height: usize,
    pub window_resizable: bool,
}

impl Default for SDLInitFlags {
    fn default() -> Self {
        Self {
            window_width: 800,
            window_height: 600,
            window_resizable: false,
        }
    }
}

pub struct SDLRenderInterface {
    sdl: sdl2::Sdl,
    video: sdl2::VideoSubsystem,
    // window: sdl2::video::Window,
    canvas: render::WindowSurface,
}

impl SDLRenderInterface {
    pub fn new(sdlif: SDLInitFlags) -> (Self, SDLEventInterface) {
        let sdl = sdl2::init().unwrap();
        let evt = sdl.event_pump().unwrap();
        let video = sdl.video().unwrap();

        let window = video
            .window(
                "title",
                sdlif.window_width as u32,
                sdlif.window_height as u32,
            )
            .position_centered()
            .build()
            .unwrap();

        let canvas = window.into_canvas().present_vsync().build().unwrap();

        (
            Self {
                sdl,
                video,
                canvas: canvas.into(),
            },
            SDLEventInterface { evt, next: None },
        )
    }
}

impl platform::RenderInterface for SDLRenderInterface {
    type RenderTarget = render::WindowSurface;
    fn render_target_mut(&mut self) -> &mut Self::RenderTarget {
        &mut self.canvas
    }
}

pub struct SDLEventInterface {
    evt: sdl2::EventPump,
    next: Option<sdl2::event::Event>,
}

impl SDLEventInterface {
    fn translate(&self, sdlevent: sdl2::event::Event) -> Option<platform::event::WindowEvent> {
        match sdlevent {
            sdl2::event::Event::Quit { .. } => Some(platform::event::WindowEvent::CloseRequest),
            _ => None,
        }
    }

    fn poll_internal(&mut self) -> Option<sdl2::event::Event> {
        self.next.take().or_else(|| self.evt.poll_event())
    }
}

impl platform::EventInterface for SDLEventInterface {
    fn poll_event(&mut self) -> Option<platform::event::WindowEvent> {
        while let Some(evt) = self.poll_internal() {
            if let Some(wevt) = self.translate(evt) {
                return Some(wevt);
            }
        }
        None
    }
    fn wait_for_event(&mut self) {
        if self.next.is_some() {
            return;
        }
        self.next = Some(self.evt.wait_event());
    }
}

pub struct SDLTextInterface {
    ctx: sdl2::ttf::Sdl2TtfContext,
}

impl platform::TextInterface for SDLTextInterface {
    fn render_text(&self, what: &str, _sz: f32) -> Box<dyn patina::platform::render::Surface + '_> {
        let font = self.ctx.load_font("/usr/share/fonts/truetype/dejavu/DejaVuSans-Bold.ttf", 16).unwrap();
        let render = font.render(what);
        let rendered = render.solid(sdl2::pixels::Color::RGB(255, 255, 255)).unwrap();
        Box::new(render::SDLSurface { surface: rendered })
    }
}

pub fn init_sdl_platform(
    sdlif: SDLInitFlags,
) -> (SDLRenderInterface, SDLEventInterface, SDLTextInterface) {
    let (sdl, event) = SDLRenderInterface::new(sdlif);
    let ttf = SDLTextInterface { ctx: sdl2::ttf::init().unwrap() };
    (sdl, event, ttf)
}
