pub type IVector = Vector<i32>;
pub type IPoint = Point<i32>;
pub type IRect = Rect<i32>;

pub type FVector = Vector<f32>;
pub type FPoint = Point<f32>;
pub type FRect = Rect<f32>;

pub mod prelude {
    pub use super::{FPoint, FRect, FVector};
    pub use super::{IPoint, IRect, IVector};
}

pub trait GeomField:
    Sized
    + Copy
    + std::ops::Add<Self, Output = Self>
    + std::ops::AddAssign<Self>
    + std::ops::Neg<Output = Self>
    + std::ops::Sub<Self, Output = Self>
    + std::ops::Mul<Self, Output = Self>
    + std::ops::MulAssign<Self>
    + PartialEq
    + PartialOrd
{
    const ZERO: Self;

    fn min(&self, other: Self) -> Self {
        if *self < other {
            *self
        } else {
            other
        }
    }

    fn max(&self, other: Self) -> Self {
        if *self > other {
            *self
        } else {
            other
        }
    }
}

impl GeomField for i32 {
    const ZERO: Self = 0;
}

impl GeomField for f32 {
    const ZERO: Self = 0.;
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd)]
pub struct Point<GF: GeomField> {
    pub x: GF,
    pub y: GF,
}

impl<GF: GeomField> Point<GF> {
    pub const ORIGIN: Self = Self {
        x: GF::ZERO,
        y: GF::ZERO,
    };
}

impl<GF: GeomField> Default for Point<GF> {
    fn default() -> Self {
        Self::ORIGIN
    }
}

impl<GF: GeomField> std::ops::Add<Vector<GF>> for Point<GF> {
    type Output = Self;
    fn add(self, rhs: Vector<GF>) -> Self::Output {
        Self {
            x: self.x + rhs.x,
            y: self.x + rhs.x,
        }
    }
}

impl<GF: GeomField> std::ops::Sub for Point<GF> {
    type Output = Vector<GF>;
    fn sub(self, rhs: Self) -> Self::Output {
        Vector {
            x: self.x - rhs.x,
            y: self.y - rhs.y,
        }
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd)]
pub struct Vector<GF: GeomField> {
    pub x: GF,
    pub y: GF,
}

impl<GF: GeomField> Vector<GF> {
    pub const ZERO: Self = Self {
        x: GF::ZERO,
        y: GF::ZERO,
    };
}

impl<GF: GeomField> std::ops::Add for Vector<GF> {
    type Output = Self;
    fn add(self, rhs: Self) -> Self::Output {
        Self {
            x: self.x + rhs.x,
            y: self.y + rhs.y,
        }
    }
}

impl<GF: GeomField> std::ops::Add<Point<GF>> for Vector<GF> {
    type Output = Point<GF>;
    fn add(self, rhs: Point<GF>) -> Self::Output {
        Point {
            x: self.x + rhs.x,
            y: self.y + rhs.y,
        }
    }
}

impl<GF: GeomField> std::ops::Sub for Vector<GF> {
    type Output = Vector<GF>;
    fn sub(self, rhs: Self) -> Self::Output {
        Vector {
            x: self.x - rhs.x,
            y: self.y - rhs.y,
        }
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd)]
pub struct Rect<GF: GeomField> {
    pub base: Point<GF>,
    pub size: Vector<GF>,
}

impl<GF: GeomField> Rect<GF> {
    pub fn new_from_size(pt: Point<GF>, sz: Vector<GF>) -> Self {
        Self { base: pt, size: sz }
    }

    pub fn new_from_points(p1: Point<GF>, p2: Point<GF>) -> Self {
        let xmin = p1.x.min(p2.x);
        let ymin = p1.y.min(p2.y);
        let xmax = p1.x.max(p2.x);
        let ymax = p1.y.max(p2.y);
        Self {
            base: Point { x: xmin, y: ymin },
            size: Vector {
                x: xmax - xmin,
                y: ymax - ymin,
            },
        }
    }

    pub fn width(&self) -> GF {
        self.size.x
    }

    pub fn height(&self) -> GF {
        self.size.y
    }

    pub fn normalize(&mut self) {
        if self.size.x < GF::ZERO {
            self.base.x += self.size.x;
            self.size.x = -self.size.x;
        }
        if self.size.y < GF::ZERO {
            self.base.y += self.size.y;
            self.size.y = -self.size.y;
        }
    }

    pub fn ll(&self) -> Point<GF> {
        self.base
    }

    pub fn ul(&self) -> Point<GF> {
        self.base
            + Vector {
                x: GF::ZERO,
                y: self.size.y,
            }
    }

    pub fn lr(&self) -> Point<GF> {
        self.base
            + Vector {
                x: self.size.x,
                y: GF::ZERO,
            }
    }

    pub fn ur(&self) -> Point<GF> {
        self.extent()
    }

    pub fn base(&self) -> Point<GF> {
        self.base
    }

    pub fn extent(&self) -> Point<GF> {
        self.base + self.size
    }

    pub fn size(&self) -> Vector<GF> {
        self.size
    }

    pub fn intersect(&self, other: &Self) -> Option<Self> {
        /*
        Some(Self {
            minpoint: Point { x: self.minpoint.x.max(other.minpoint.x), y: self.minpoint.y.max(other.minpoint.y) },
            maxpoint: Point { x: self.maxpoint.x.min(other.maxpoint.x), y: self.maxpoint.y.min(other.maxpoint.y) },
        })
        */
        None
    }

    pub fn contains(&self, pt: Point<GF>) -> bool {
        (self.base.x <= pt.x)
            && (pt.x < self.extent().x)
            && (self.base.y <= pt.y)
            && (pt.y < self.extent().y)
    }
}
